#version 430

layout(location = 0) in vec3 position;
layout(location = 1) uniform vec3 color;
layout(location = 2) in vec3 normal;
layout(location = 3) uniform float opacity;
layout(location = 4) uniform float ambientPower;

layout(location = 5) uniform mat4 vp;
layout(location = 15) uniform mat4 model;

layout(location = 20) uniform float ka;
layout(location = 21) uniform float kd;
layout(location = 22) uniform float ks;
layout(location = 23) uniform int alpha;

struct DirectionalLight {
    vec3 direction;
    vec3 color;
};  
layout(location = 30) uniform DirectionalLight directionals[3];

layout(location = 40) uniform vec3 cameraPosition;
layout(location = 41) uniform mat4 light_vp;
layout(location = 45) uniform sampler2D shadowMap;

struct PointLight {
    vec3 color;
    vec3 position;

    float attenuationConstant;
    float attenuationLinear;
    float attenuationQuadratic;
};
layout(location = 50) uniform PointLight pointLights[3];

struct SpotLight {
    vec3 color;
    vec3 position;
    vec3 direction;

    float angleInner;
    float angleOuter;

    vec3 attenuation; //attenuationConstant, attenuationLinear, attenuationQuadratic;

    bool shadows;
    mat4 light_vp;
};
layout(location = 70) uniform SpotLight spotLights[3];

layout(location = 100) uniform int numDirectional;
layout(location = 101) uniform int numPoint;
layout(location = 102) uniform int numSpot;

out vec3 position_world;
out vec3 normal_world;
out vec4 gouraud_color;


vec3 calculateDiffuse(vec3 lightVector, vec3 lightColor) {
    float cosTheta = clamp(dot(normalize(normal_world),lightVector), 0,1 ); // = n dot l
    return (kd * color * lightColor * cosTheta);
}

vec3 calculateSpecular(vec3 lightVector, vec3 viewVector, vec3 lightColor) {
    vec3 r = reflect(-lightVector, normalize(normal_world));
    float cosPhi = clamp(dot(viewVector, r), 0,1 );
    return (ks * pow(cosPhi,alpha) * lightColor);
}

vec3 calculatePointLight(PointLight point, vec3 viewVector){
    vec3 lPoint = point.position - position_world;
    float dist = length(lPoint);
    lPoint = normalize(lPoint);
    float fAtt = 1/(point.attenuationConstant + dist * point.attenuationLinear + dist * dist * point.attenuationQuadratic);

    return calculateDiffuse(lPoint, point.color) * fAtt 
        + calculateSpecular(lPoint, viewVector, point.color) * fAtt;
}

vec3 calculateSpotLight(SpotLight spot, vec3 viewVector){
    vec3 lSpot = spot.position - position_world;
    float dist = length(lSpot);
    lSpot = normalize(lSpot);

    float cosAngleInCone = dot(normalize(-spot.direction), lSpot);
    float angleInCone = acos(cosAngleInCone);

    if(angleInCone < spot.angleOuter){
            
        float falloff = 1;
        if(angleInCone >= spot.angleInner){
            falloff = 1-smoothstep(spot.angleInner, spot.angleOuter, angleInCone);
        }

        float attenuationConstant = spot.attenuation.x;
        float attenuationLinear = spot.attenuation.y;
        float attenuationQuadratic = spot.attenuation.z;

        float fAttSpot = 1/(attenuationConstant + dist * attenuationLinear + dist * dist * attenuationQuadratic);

        return calculateDiffuse(lSpot, spot.color) * fAttSpot * falloff
            + calculateSpecular(lSpot, viewVector, spot.color) * fAttSpot * falloff;
    }
    return vec3(0);
}

vec3 calculateDirectional(DirectionalLight directionalLight, vec3 viewVector){
    vec3 lDirectional = normalize(-directionalLight.direction);
    return calculateDiffuse(lDirectional, directionalLight.color)
        + calculateSpecular(lDirectional, viewVector, directionalLight.color);
 }

vec4 calculateLights(){
        vec3 ambient = ambientPower * ka * color;
        vec3 result = ambient;

        vec3 v = normalize(cameraPosition-position_world);

        //directional
        for(int i = 0; i < numDirectional; i++)
            result += calculateDirectional(directionals[i], v);
        
        //point
        for(int i = 0; i < numPoint; i++)
            result += calculatePointLight(pointLights[i], v);

        //spot
        for(int i = 0; i < numSpot; i++)
            result += calculateSpotLight(spotLights[i], v);

        return vec4(result, 1);
}

void main(){
    gl_Position = vp * model * vec4(position, 1);

    position_world = (model * vec4(position, 1)).xyz; 
    normal_world = (transpose(inverse(model)) * vec4(normal, 1)).xyz;

    gouraud_color = calculateLights();
    gouraud_color.a = opacity;
}